from nsqio.http.base import NsqHTTPConnection
from nsqio.utils import _convert_to_str


class NsqdHttpWriter(NsqHTTPConnection):
    """
    :see: http://nsq.io/components/nsqd.html
    """

    async def ping(self):
        """Monitoring endpoint.
        :returns: should return `"OK"`, otherwise raises an exception.
        """
        return await self.perform_request("GET", "ping", None, None)

    async def info(self):
        """Returns version information."""
        resp = await self.perform_request("GET", "info", None, None)
        return resp

    async def stats(self):
        """Returns version information."""
        resp = await self.perform_request("GET", "stats", {"format": "json"}, None)
        return resp

    async def pub(self, topic, message):
        """Returns version information."""
        resp = await self.perform_request("POST", "pub", {"topic": topic}, message)
        return resp

    async def mpub(self, topic, *messages):
        """Returns version information."""
        assert len(messages), "Specify one or mor message"
        _msgs = [_convert_to_str(m) for m in messages]
        msgs = "\n".join(_msgs)
        resp = await self.perform_request("POST", "mpub", {"topic": topic}, msgs)
        return resp

    async def create_topic(self, topic):
        resp = await self.perform_request(
            "POST", "topic/create", {"topic": topic}, None
        )
        return resp

    async def delete_topic(self, topic):
        resp = await self.perform_request(
            "POST", "topic/delete", {"topic": topic}, None
        )
        return resp

    async def create_channel(self, topic, channel):
        resp = await self.perform_request(
            "POST", "channel/create", {"topic": topic, "channel": channel}, None
        )
        return resp

    async def delete_channel(self, topic, channel):
        resp = await self.perform_request(
            "POST", "channel/delete", {"topic": topic, "channel": channel}, None
        )
        return resp

    async def empty_topic(self, topic):
        resp = await self.perform_request("POST", "topic/empty", {"topic": topic}, None)
        return resp

    async def topic_pause(self, topic):
        resp = await self.perform_request("POST", "topic/pause", {"topic": topic}, None)
        return resp

    async def topic_unpause(self, topic):
        resp = await self.perform_request(
            "POST", "topic/unpause", {"topic": topic}, None
        )
        return resp

    async def pause_channel(self, channel, topic):
        resp = await self.perform_request(
            "POST", "channel/pause", {"topic": topic, "channel": channel}, None
        )
        return resp

    async def unpause_channel(self, channel, topic):
        resp = await self.perform_request(
            "POST", "channel/unpause", {"topic": topic, "channel": channel}, None
        )
        return resp

    async def debug_pprof(self):
        resp = await self.perform_request("GET", "debug/pprof", None, None)
        return resp

    async def debug_pprof_profile(self):
        resp = await self.perform_request("GET", "debug/pprof/profile", None, None)
        return resp

    async def debug_pprof_goroutine(self):
        resp = await self.perform_request("GET", "debug/pprof/goroutine", None, None)
        return resp

    async def debug_pprof_heap(self):
        resp = await self.perform_request("GET", "debug/pprof/heap", None, None)
        return resp

    async def debug_pprof_block(self):
        resp = await self.perform_request("GET", "debug/pprof/block", None, None)
        return resp

    async def debug_pprof_threadcreate(self):
        resp = await self.perform_request("GET", "debug/pprof/threadcreate", None, None)
        return resp

    async def nsqlookupd_tcp_addresses(self):
        """
        List of nsqlookupd TCP addresses.
        """
        resp = await self.perform_request(
            "GET", "config/nsqlookupd_tcp_addresses", None, None
        )
        return resp



"""NSQ protocol parser.

:see: http://nsq.io/clients/tcp_protocol_spec.html
"""
import abc
import struct
import zlib
import snappy

from nsqio.tcp.consts import (
    DATA_SIZE,
    FRAME_SIZE,
    FRAME_TYPE_RESPONSE,
    FRAME_TYPE_ERROR,
    FRAME_TYPE_MESSAGE,
    MSG_HEADER,
    NL,
)

from nsqio.tcp.exceptions import ProtocolError
from nsqio.utils import _convert_to_bytes, get_logger

logger = get_logger()


__all__ = ["Reader", "DeflateReader", "SnappyReader"]


class BaseReader(metaclass=abc.ABCMeta):
    @abc.abstractmethod  # pragma: no cover
    def feed(self, chunk):
        """

        :return:
        """

    @abc.abstractmethod  # pragma: no cover
    def gets(self):
        """

        :return:
        """

    @abc.abstractmethod  # pragma: no cover
    def encode_command(self, cmd, *args, data=None):
        """

        :return:
        """


class BaseCompressReader(BaseReader):
    @abc.abstractmethod  # pragma: no cover
    def compress(self, data):
        """

        :param data:
        :return:
        """

    @abc.abstractmethod  # pragma: no cover
    def decompress(self, chunk):
        """

        :param chunk:
        :return:
        """

    def feed(self, chunk):
        if not chunk:
            return
        uncompressed = self.decompress(chunk)
        uncompressed and self._parser.feed(uncompressed)

    def gets(self):
        return self._parser.gets()

    def encode_command(self, cmd, *args, data=None):
        cmd = self._parser.encode_command(cmd, *args, data=data)
        # print(cmd)
        return self.compress(cmd)


class DeflateReader(BaseCompressReader):
    def __init__(self, buffer=None, level=6):
        self._parser = Reader()
        wbits = -zlib.MAX_WBITS
        self._decompressor = zlib.decompressobj(wbits)
        self._compressor = zlib.compressobj(level, zlib.DEFLATED, wbits)
        buffer and self.feed(buffer)

    def compress(self, data):
        chunk = self._compressor.compress(data)
        compressed = chunk + self._compressor.flush(zlib.Z_SYNC_FLUSH)
        return compressed

    def decompress(self, chunk):
        return self._decompressor.decompress(chunk)


class SnappyReader(BaseCompressReader):
    def __init__(self, buffer=None):
        self._parser = Reader()
        self._decompressor = snappy.StreamDecompressor()
        self._compressor = snappy.StreamCompressor()
        buffer and self.feed(buffer)

    def compress(self, data):
        compressed = self._compressor.add_chunk(data, compress=True)
        return compressed

    def decompress(self, chunk):
        return self._decompressor.decompress(chunk)


def _encode_body(data):
    _data = _convert_to_bytes(data)
    result = struct.pack(">l", len(_data)) + _data
    return result


class Reader(BaseReader):
    def __init__(self, buffer=None):

        self._buffer = bytearray()
        self._payload_size = None
        self._is_header = False
        self._frame_type = None
        buffer and self.feed(buffer)

    @property
    def buffer(self):
        return self._buffer

    def feed(self, chunk):
        """Put raw chunk of data obtained from connection to buffer.
        :param data: ``bytes``, raw input data.
        """
        if not chunk:
            return
        self._buffer.extend(chunk)

    def gets(self):
        buffer_size = len(self._buffer)
        if not self._is_header and buffer_size >= DATA_SIZE:
            size = struct.unpack(">l", self._buffer[:DATA_SIZE])[0]
            self._payload_size = size
            self._is_header = True

        if self._is_header and buffer_size >= DATA_SIZE + self._payload_size:

            start, end = DATA_SIZE, DATA_SIZE + FRAME_SIZE

            self._frame_type = struct.unpack(">l", self._buffer[start:end])[0]
            # temp neglect for frame error.
            # todo
            if self._frame_type not in (
                FRAME_TYPE_RESPONSE,
                FRAME_TYPE_ERROR,
                FRAME_TYPE_MESSAGE,
            ):
                logger.debug(f"_frame_type error-> {self._frame_type}")
                self._reset()
                return False
            resp = self._parse_payload()
            self._reset()
            return resp
        return False

    def _reset(self):
        start = DATA_SIZE + self._payload_size
        self._buffer = self._buffer[start:]
        self._is_header = False
        self._payload_size = None
        self._frame_type = None

    def _parse_payload(self):

        response_type, response = self._frame_type, None
        if response_type == FRAME_TYPE_RESPONSE:
            response = self._unpack_response()
        elif response_type == FRAME_TYPE_ERROR:
            response = self._unpack_error()
        elif response_type == FRAME_TYPE_MESSAGE:
            response = self._unpack_message()
        else:
            raise ProtocolError()
        return response_type, response

    def _unpack_error(self):
        start = DATA_SIZE + FRAME_SIZE
        end = DATA_SIZE + self._payload_size
        error = bytes(self._buffer[start:end])
        code, msg = error.split(None, 1)
        return code, msg

    def _unpack_response(self):
        start = DATA_SIZE + FRAME_SIZE
        end = DATA_SIZE + self._payload_size
        body = bytes(self._buffer[start:end])
        return body

    def _unpack_message(self):
        start = DATA_SIZE + FRAME_SIZE
        end = DATA_SIZE + self._payload_size
        msg_len = end - start - MSG_HEADER
        fmt = ">qh16s{}s".format(msg_len)
        payload = struct.unpack(fmt, self._buffer[start:end])
        timestamp, attempts, msg_id, body = payload
        return timestamp, attempts, msg_id, body

    def encode_command(self, cmd, *args, data=None):
        """XXX"""
        _cmd = _convert_to_bytes(cmd.upper().strip())
        _args = [_convert_to_bytes(a) for a in args]
        body_data, params_data = b"", b""

        if len(_args):
            params_data = b" " + b" ".join(_args)

        if data and isinstance(data, (list, tuple)):
            data_encoded = [_encode_body(part) for part in data]
            num_parts = len(data_encoded)
            payload = struct.pack(">l", num_parts) + b"".join(data_encoded)
            body_data = struct.pack(">l", len(payload)) + payload
        elif data:
            body_data = _encode_body(data)

        return b"".join((_cmd, params_data, NL, body_data))




class NSQException(Exception):
    """XXX"""


class ProtocolError(NSQException):
    """XXX"""


class NSQRequeueMessage(NSQException):
    pass


class NSQNoConnections(NSQException):
    pass


class NSQHttpError(NSQException):
    pass


class NSQFrameError(NSQException):
    pass


class NSQErrorCode(NSQException):
    fatal = True


class NSQInvalid(NSQErrorCode):
    """E_INVALID"""
    pass


class NSQBadBody(NSQErrorCode):
    """E_BAD_BODY"""
    pass


class NSQBadTopic(NSQErrorCode):
    """E_BAD_TOPIC"""
    pass


class NSQBadChannel(NSQErrorCode):
    """E_BAD_CHANNEL"""
    pass


class NSQBadMessage(NSQErrorCode):
    """E_BAD_MESSAGE"""
    pass


class NSQPutFailed(NSQErrorCode):
    """E_PUT_FAILED"""
    pass


class NSQPubFailed(NSQErrorCode):
    """E_PUB_FAILED"""


class NSQMPubFailed(NSQErrorCode):
    """E_MPUB_FAILED"""


class NSQAuthDisabled(NSQErrorCode):
    """E_AUTH_DISABLED"""


class NSQAuthFailed(NSQErrorCode):
    """E_AUTH_FAILED"""


class NSQUnauthorized(NSQErrorCode):
    """E_UNAUTHORIZED"""


class NSQFinishFailed(NSQErrorCode):
    """E_FIN_FAILED"""
    fatal = False


class NSQRequeueFailed(NSQErrorCode):
    """E_REQ_FAILED"""
    fatal = False


class NSQTouchFailed(NSQErrorCode):
    """E_TOUCH_FAILED"""
    fatal = False


ERROR_CODES = {
    b'E_INVALID': NSQInvalid,
    b'E_BAD_BODY': NSQBadBody,
    b'E_BAD_TOPIC': NSQBadTopic,
    b'E_BAD_CHANNEL': NSQBadChannel,
    b'E_BAD_MESSAGE': NSQBadMessage,
    b'E_PUT_FAILED': NSQPutFailed,
    b'E_PUB_FAILED': NSQPubFailed,
    b'E_MPUB_FAILED': NSQMPubFailed,
    b'E_FINISH_FAILED': NSQFinishFailed,
    b'E_AUTH_DISABLED': NSQAuthDisabled,
    b'E_AUTH_FAILED': NSQAuthFailed,
    b'E_UNAUTHORIZED': NSQUnauthorized,
    b'E_FIN_FAILED': NSQFinishFailed,
    b'E_REQUEUE_FAILED': NSQRequeueFailed,
    b'E_REQ_FAILED': NSQRequeueFailed,
    b'E_TOUCH_FAILED': NSQTouchFailed
}

# https://groups.google.com/forum/#!msg/nsq-users/VSxdPtw2ZmY/kmQJJhZe4wEJ
# E_TOUCH_FAILED, fatal
# E_REQ_FAILED fatal
# E_FIN_FAILED


def make_error(code, error_message):
    return ERROR_CODES.get(code, NSQErrorCode)(error_message)



from collections import namedtuple
from nsqio.tcp.consts import TOUCH, REQ, FIN
from nsqio.utils import get_logger


__all__ = ["NsqMessage", "NsqErrorMessage"]


NsqErrorMessage = namedtuple("NsqError", ["code", "msg"])
BaseMessage = namedtuple("NsqMessage", "timestamp attempts message_id body conn")

logger = get_logger()


class NsqMessage(BaseMessage):
    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls, *args, **kwargs)
        self._is_processed = False
        return self

    @property
    def processed(self):
        """True if message has been processed: finished or re-queued."""
        return self._is_processed

    async def fin(self):
        """Finish a message (indicate successful processing)

        """
        if self._is_processed:
            logger.warning("{} has already been processed".format(self))
            return None
        resp = await self.conn.execute(FIN, self.message_id)
        self._is_processed = True
        return resp

    async def req(self, timeout=10):
        """Re-queue a message (indicate failure to process)

        :param timeout: ``int`` configured max timeout  0 is a special case
            that will not defer re-queueing.

        """
        if self._is_processed:
            logger.warning("{} has already been processed".format(self))
            return None
        resp = await self.conn.execute(REQ, self.message_id, timeout)
        self._is_processed = True
        return resp

    async def touch(self):
        """Reset the timeout for an in-flight message.
        """
        if self._is_processed:
            logger.warning("{} has already been processed".format(self))
            return None
        return await self.conn.execute(TOUCH, self.message_id)

    def __repr__(self):
        return "<NsqMessage{}@{}>".format(self.message_id, self.conn)



import random
import re
import logzero
import logging
from urllib.parse import urlparse

TOPIC_NAME_RE = re.compile(r"^[\.a-zA-Z0-9_-]+$")
CHANNEL_NAME_RE = re.compile(r"^[\.a-zA-Z0-9_-]+(#ephemeral)?$")


def get_host_and_port(host):
    host_parsed = urlparse(host)
    if host_parsed.scheme == "tcp":
        result = host_parsed.netloc
    elif host_parsed.scheme == "":
        result = host_parsed.path
    else:
        result = host
    result = result.split(":")
    if len(result) == 2:
        return result[0], result[-1]
    else:
        return result[0], None


def get_version():
    from nsqio import __version__

    return __version__


def valid_topic_name(topic):
    if not 0 < len(topic) < 33:
        return False
    return bool(TOPIC_NAME_RE.match(topic))


def valid_channel_name(channel):
    if not 0 < len(channel) < 33:
        return False
    return bool(CHANNEL_NAME_RE.match(channel))


_converters_to_bytes_map = {
    bytes: lambda val: val,
    bytearray: lambda val: val,
    str: lambda val: val.encode("utf-8"),
    int: lambda val: str(val).encode("utf-8"),
    float: lambda val: str(val).encode("utf-8"),
}


_converters_to_str_map = {
    str: lambda val: val,
    bytearray: lambda val: bytes(val).decode("utf-8"),
    bytes: lambda val: val.decode("utf-8"),
    int: lambda val: str(val),
    float: lambda val: str(val),
}


def _convert_to_bytes(value):
    if type(value) in _converters_to_bytes_map:
        converted_value = _converters_to_bytes_map[type(value)](value)
    else:
        raise TypeError(
            "Argument {!r} expected to be of bytes,"
            " str, int or float type".format(value)
        )
    return converted_value


def _convert_to_str(value):
    if type(value) in _converters_to_str_map:
        converted_value = _converters_to_str_map[type(value)](value)
    else:
        raise TypeError(
            "Argument {!r} expected to be of bytes,"
            " str, int or float type".format(value)
        )
    return converted_value


class MaxRetriesExided(Exception):
    pass


def retry_iterator(
    init_delay=0.1,
    max_delay=10.0,
    factor=2.7182818284590451,
    jitter=0.11962656472,
    max_retries=None,
    now=True,
):
    """Based on twisted reconnection factory.

    :param init_delay:
    :param max_delay:
    :param factor:
    :param jitter:
    :param max_retries:
    :param now:
    :return:
    """
    retries, delay = 0, init_delay
    if now:
        retries += 1
        yield 0.0
    while not max_retries or retries < max_retries:
        retries += 1
        delay *= factor
        delay = random.normalvariate(delay, delay * jitter) if jitter else delay
        delay = min(delay, max_delay) if max_delay else delay
        yield delay
    else:
        raise MaxRetriesExided()


_logger = None


def setup_logger(
    logfile=None,
    level=logging.DEBUG,
    formatter=None,
    maxBytes=0,
    backupCount=0,
    fileLoglevel=None,
    disableStderrLogger=False,
):
    global _logger
    _logger = logzero.setup_logger(
        "nsqio",
        logfile=logfile,
        level=level,
        formatter=formatter,
        maxBytes=maxBytes,
        backupCount=backupCount,
        fileLoglevel=fileLoglevel,
        disableStderrLogger=disableStderrLogger,
    )
    return _logger


# initialize logger
def _check_and_init_logger():
    if _logger is None:
        setup_logger()


_check_and_init_logger()


def get_logger():
    return _logger



from nsqio.tcp.connection import logger
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from nsqio.tcp.connection import TcpConnection
    from typing import Dict

import asyncio
import random
import traceback
from nsqio.tcp.consts import RDY

REDISTRIBUTE = 0
CHANGE_CONN_RDY = 1
NOOP = 2


class RdyControl:
    def __init__(self, idle_timeout, max_in_flight, loop=None):
        self._connections: "Dict[str, TcpConnection]" = {}
        self._idle_timeout = idle_timeout
        self._total_ready_count = 0
        self._max_in_flight = max_in_flight
        self._loop = loop or asyncio.get_event_loop()

        self._cmd_queue = asyncio.Queue(loop=self._loop)

        self._expected_rdy_state = {}

        self._is_working = True

        self._distributor_task = self._loop.create_task(self._distributor())

    @property
    def connections(self):
        return self._connections

    def add_connections(self, connections):
        prev_connections = self._connections
        self._connections = connections
        for conn in self._connections.values():
            conn._on_rdy_changed_cb = self.rdy_changed

        self._close_all_connections(prev_connections)

    def add_connection(self, connection):
        connection._on_rdy_changed_cb = self.rdy_changed
        id = connection.id
        if id in self._connections:
            self._connections[id].close()
        self._connections[connection.id] = connection

    def rdy_changed(self, conn_id):
        self._cmd_queue.put_nowait((CHANGE_CONN_RDY, (conn_id,)))

    def redistribute(self):
        self._cmd_queue.put_nowait((REDISTRIBUTE, ()))

    async def _distributor(self):
        while self._is_working:
            cmd, args = await self._cmd_queue.get()
            if cmd == REDISTRIBUTE:
                await self._redistribute_rdy_state()
            elif cmd == CHANGE_CONN_RDY:
                await self._update_rdy(*args)
            elif cmd == NOOP:
                continue
            else:
                RuntimeError("Should never be here")

    def remove_connection(self, conn: "TcpConnection"):
        self._connections.pop(conn.id)
        conn.close()

    def remove_all(self):
        prev_connections = self._connections

        self._connections = {}

        self._close_all_connections(prev_connections)

    def _close_all_connections(self, connections):
        for conn in connections.values():
            if not conn.closed:
                conn.close()

    def stop_working(self):
        self._is_working = False
        self._cmd_queue.put_nowait((NOOP, ()))
        self.remove_all()

    async def _redistribute_rdy_state(self):
        # We redistribute RDY counts in a few cases:
        #
        # 1. our # of connections exceeds our configured max_in_flight
        # 2. we're in backoff mode (but not in a current backoff block)
        # 3. something out-of-band has set the need_rdy_redistributed flag
        # (connection closed
        # that was about to get RDY during backoff)
        #
        # At a high level, we're trying to mitigate stalls related to
        # -volume
        # producers when we're unable (by configuration or backoff) to provide
        # a RDY count
        # of (at least) 1 to all of our connections.

        # check all connections
        for conn_id in set(self._connections.keys()):
            await self._is_valid_connection(conn_id)

        if len(self._connections.keys()) == 0:
            logger.warning("no valid connectionss.., skip...")
            return

        connections = self._connections.values()
        # disable for further deprecate

        # rdy_coros = [
        #     conn.execute(RDY, 0) for conn in connections
        #     if not (conn.rdy_state == 0 or
        #             (time.time() - conn.last_message) < self._idle_timeout)
        # ]

        distributed_rdy = sum(c._in_flight for c in connections)
        not_distributed_rdy = self._max_in_flight - distributed_rdy

        random_connections = random.sample(
            list(connections), min(not_distributed_rdy, len(connections))
        )

        rdy_coros = [conn.execute(RDY, 1) for conn in random_connections]

        await asyncio.gather(*rdy_coros, loop=self._loop)

    async def _is_valid_connection(self, conn_id):
        conn = self._connections.get(conn_id, None)
        if not conn:
            logger.debug("connection {} not found, skip...".format(conn_id))
            return False

        if conn.closed:
            try:
                logger.warning("connection {} closed, removing..".format(conn))
                await self.remove_connection(conn)
            except Exception as e:
                logger.error(
                    "remove_connectionfailed: {}\n{}".format(e, traceback.format_exc())
                )
            finally:
                return False
        return True

    async def _update_rdy(self, conn_id):
        logger.debug("_update_rdy :{}".format(conn_id))
        if not await self._is_valid_connection(conn_id):
            return

        conn = self._connections[conn_id]

        # this is the configuration max_in_flight split even on conn
        base_conn_max_in_flight = self._max_in_flight / max(1, len(self._connections))

        # this is the in_flight number of the conn_id's conn
        conn_in_flight = conn._in_flight

        # get the max rdy state for conn
        rdy_state = int(max(1, base_conn_max_in_flight - conn_in_flight))
        await conn.execute(RDY, rdy_state)



import re
import os.path
import sys
from setuptools import setup, find_packages


install_requires = [
    "python-snappy>=0.5.4",
    "aiohttp>=3.6.0",
    "logzero>=1.5.0",
]

NAME = "nsqio"
PACKAGE = "nsqio"
PY_VER = sys.version_info

if PY_VER >= (3, 6):
    pass
else:
    raise RuntimeError("nsqio doesn't support Python version prior 3.6")


def read(*parts):
    with open(os.path.join(*parts), "rt") as f:
        return f.read().strip()


def read_version():
    regexp = re.compile(r"^__version__\W*=\W*\"([\d.abrc]+)\"")
    init_py = os.path.join(os.path.dirname(__file__), "nsqio", "__init__.py")
    with open(init_py) as f:
        for line in f:
            match = regexp.match(line)
            if match is not None:
                return match.group(1)
        else:
            raise RuntimeError("Cannot find version in nsqio/__init__.py")


classifiers = [
    "License :: OSI Approved :: MIT License",
    "Development Status :: 3 - Alpha",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Operating System :: POSIX",
    "Environment :: Web Environment",
    "Intended Audience :: Developers",
    "Topic :: Software Development",
    "Topic :: Software Development :: Libraries",
]


if os.path.exists("README.md"):
    with open("README.md", "r") as f:
        long_description = f.read()
else:
    long_description = "See http://pypi.python.org/pypi/%s" % (NAME,)

setup(
    name="nsqio",
    version=read_version(),
    description=("asyncio async/await nsq support"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=classifiers,
    platforms=["POSIX"],
    author="yuikns",
    author_email="yuikns@users.noreply.github.com",
    url="https://github.com/rcrai/nsqio",
    license="MIT",
    packages=find_packages(exclude=["tests"]),
    install_requires=install_requires,
    include_package_data=True,
)



import asyncio
import json
import ssl

# import logging

from collections import deque

from nsqio.tcp.consts import (
    MAGIC_V2,
    BIN_OK,
    HEARTBEAT,
    FRAME_TYPE_RESPONSE,
    FRAME_TYPE_ERROR,
    FRAME_TYPE_MESSAGE,
)

from nsqio.tcp.messages import NsqMessage
from nsqio.utils import get_logger
from nsqio.tcp.exceptions import ProtocolError  # , make_error
from nsqio.tcp.protocol import Reader, DeflateReader, SnappyReader

logger = get_logger()


async def create_connection(host="localhost", port=4150, queue=None, loop=None):
    """create nsq tcp connection
    Args:
        host: host address
        port: host port
        queue: user define asyncio queue
        loop: user define asyncio loop
    Return:
        TcpConnection
    """
    reader, writer = await asyncio.open_connection(host, port, loop=loop)
    conn = TcpConnection(reader, writer, host, port, queue=queue, loop=loop)
    conn.connect()
    return conn


class TcpConnection:
    """
    base nsq connection class ,used for manipulate reader/writer content
    """

    def __init__(
        self,
        reader,
        writer,
        host,
        port,
        *,
        on_rdy_changed=None,
        on_message=None,
        queue=None,
        loop=None,
    ):
        self._reader, self._writer = reader, writer
        self._host, self._port = host, port

        self._loop = loop or asyncio.get_event_loop()

        assert isinstance(queue, asyncio.Queue) or queue is None
        self._queue = queue or asyncio.Queue(loop=self._loop)

        self._parser = Reader()
        # next queue is used for nsq commands
        self._cmd_waiters = deque()
        self._closing = False
        self._closed = False
        self._reader_task = self._loop.create_task(self._read_data())
        # mark connection in upgrading state to ssl socket
        self._is_upgrading = False
        self._on_message = on_message
        self._on_rdy_changed_cb = on_rdy_changed
        self._on_close = None
        self._on_close_flag = asyncio.Event(loop=self._loop)

        # number of received but not acked or req messages
        self._in_flight = 0
        logger.info("new connection: {}:{}".format(self._host, self._port))

    def connect(self):
        self._send_magic()
        logger.info("connect: {}:{}".format(self._host, self._port))

    def execute(self, command: bytes, *args, data=None, cb=None):
        """XXX"""
        assert (
            self._reader and not self._reader.at_eof()
        ), "Connection closed or corrupted"
        if command is None:
            raise TypeError("command must not be None")
        if None in set(args):
            raise TypeError("args must not contain None")
        fut = asyncio.Future(loop=self._loop)

        if command in (b"NOP", b"FIN", b"RDY", b"REQ", b"TOUCH"):
            fut.set_result(b"OK")
        else:
            self._cmd_waiters.append((fut, cb))

        command_raw = self._parser.encode_command(command, *args, data=data)
        logger.debug("execute command %s" % command_raw)
        self._writer.write(command_raw)

        # track all processed and requeued messages
        if command in (b"FIN", b"REQ", "FIN", "REQ"):
            self._in_flight = max(0, self._in_flight - 1)
        return fut

    @property
    def in_flight(self):
        return self._in_flight

    @property
    def endpoint(self):
        return "tcp://{}:{}".format(self._host, self._port)

    @property
    def id(self):
        return self.endpoint

    @property
    def closed(self):
        """True if connection is closed."""
        closed = self._closing or self._closed
        if not closed and self._reader and self._reader.at_eof():
            self._closing = closed = True
            self._loop.call_soon(self._do_close, None)
        return closed

    async def wait_for_closed(self, timeout=10):
        await asyncio.wait_for(self._on_close_flag.wait(), timeout, loop=self._loop)

    @property
    def queue(self):
        return self._queue

    def close(self):
        """Close connection."""
        logger.debug("closing {}".format(self.id))
        self._do_close()

    async def identify(self, **config):
        # TODO: add config validator
        data = json.dumps(config)
        resp = await self.execute(b"IDENTIFY", data=data, cb=self._start_upgrading)
        if resp in (b"OK", "OK"):
            self._finish_upgrading()
            return resp
        resp_config = json.loads(resp.decode("utf-8"))
        fut = None
        if resp_config.get("tls_v1"):
            await self._upgrade_to_tls()

        if resp_config.get("snappy"):
            fut = self._upgrade_to_snappy()
        elif resp_config.get("deflate"):
            fut = self._upgrade_to_deflate()
        self._finish_upgrading()
        if fut:
            ok = await fut
            assert ok == b"OK"
        return resp

    def _do_close(self, exc=None):
        logger.info("this is the going close info")
        if exc:
            logger.error("Connection closed with error: {}".format(exc))
        if self._closed:
            return
        self._closed = True
        self._closing = False
        self._writer.transport.close()
        self._reader_task.cancel()
        self._on_close_flag.set()

    def _send_magic(self):
        self._writer.write(MAGIC_V2)

    def _pulse(self):
        nop = self._parser.encode_command(b"NOP")
        self._writer.write(nop)

    async def _upgrade_to_tls(self):
        self._reader_task.cancel()
        transport = self._writer.transport
        transport.pause_reading()
        raw_sock = transport.get_extra_info("socket", default=None)
        if raw_sock is None:
            raise RuntimeError("Transport does not expose socket instance")
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1)

        self._reader, self._writer = await asyncio.open_connection(
            sock=raw_sock, ssl=ssl_context, loop=self._loop, server_hostname=self._host
        )
        bin_ok = await self._reader.readexactly(10)
        if bin_ok != BIN_OK:
            raise RuntimeError("Upgrade to TLS failed, got: {}".format(bin_ok))
        self._reader_task = self._loop.create_task(self._read_data())
        self._reader_task.add_done_callback(self._on_reader_task_stopped)

    def _on_reader_task_stopped(self, future):
        exc = future.exception()
        logger.error("DONE: TASK {}".format(exc))

    def _upgrade_to_snappy(self):
        self._parser = SnappyReader(self._parser.buffer)
        fut = asyncio.Future(loop=self._loop)
        self._cmd_waiters.append((fut, None))
        return fut

    def _upgrade_to_deflate(self):
        self._parser = DeflateReader(self._parser.buffer)
        fut = asyncio.Future(loop=self._loop)
        self._cmd_waiters.append((fut, None))
        return fut

    async def _read_data(self):
        """Response reader task."""
        is_canceled = False
        # logger.debug("{} starting _read_data".format(self))
        while not self._reader.at_eof():
            try:
                data = await self._reader.read(52)
            except asyncio.CancelledError:
                is_canceled = True
                logger.debug("Task is canceled {}".format(self))
                break
            except Exception as exc:
                logger.exception(exc)
                logger.debug("Reader task stopped due to: {}".format(exc))
                break
            self._parser.feed(data)
            not self._is_upgrading and self._read_buffer()

        if is_canceled:
            # useful during update to TLS, task canceled but connection
            # should not be closed
            return
        logger.info("{} is read to end, going to close".format(self.id))
        self._closing = True
        self._loop.call_soon(self._do_close, None)

    def _parse_data(self):
        try:
            obj = self._parser.gets()
        except ProtocolError as exc:
            # ProtocolError is fatal
            # so connection must be closed
            logger.exception(exc)
            self._closing = True
            self._loop.call_soon(self._do_close, exc)
            logger.error("ProtocolError is fatal")
            return
        else:
            if obj is False:
                return False
            logger.debug("got nsq data: %s", obj)
            resp_type, resp = obj
            hb = HEARTBEAT
            # print(resp_type, resp)
            if resp_type == FRAME_TYPE_RESPONSE and resp == hb:
                self._pulse()
            elif resp_type == FRAME_TYPE_RESPONSE:
                waiter, cb = self._cmd_waiters.popleft()
                if not waiter.cancelled():
                    waiter.set_result(resp)
                    cb is not None and cb(resp)
            elif resp_type == FRAME_TYPE_ERROR:
                waiter, cb = self._cmd_waiters.popleft()
                # error = make_error(*resp)
                if not waiter.cancelled():
                    waiter.set_result(resp)
                    cb is not None and cb(resp)
            elif resp_type == FRAME_TYPE_MESSAGE:

                # track number in flight messages
                self._in_flight += 1

                ts, att, msg_id, body = resp
                self._on_message_hook(ts, att, msg_id, body)
                # self._queue.put_nowait(msg)
            return True

    def _on_message_hook(self, ts, att, msg_id, body):
        msg = NsqMessage(ts, att, msg_id, body, self)
        if self._on_message:
            msg = self._on_message(msg)
        self._queue.put_nowait(msg)

    def _read_buffer(self):
        is_continue = True
        while is_continue:
            is_continue = self._parse_data()

    def _start_upgrading(self, resp=None):
        self._is_upgrading = True

    def _finish_upgrading(self, resp=None):
        self._read_buffer()
        self._is_upgrading = False

    def __repr__(self):
        return "<TcpConnection: {}:{}>".format(self._host, self._port)
